# External Modules
from os             import environ
from argparse       import ArgumentParser
from pprint         import pformat
from distutils.util import strtobool
from ast            import literal_eval
from json           import load

# Internal Modules
from dbgen.support.utils          import ConnectInfo
from dbgen.core.misc              import levenshteinDistance
from dbgen.inputs.development     import dbg
"""
Command line interface for executing a database update

Requires DB_JSON environment variable to specify database that will be modified

The following flags control program behavior:
    - verbose
    - local
    - reset
    - add
    - meta
    - only
    - xclude
    - jobs
    - catalog
    - test
    - parallel

"""
################################################################################
##########
# Settings
#---------
parser = ArgumentParser(description  = 'Run a DBG update'
                       ,allow_abbrev = True)

parser.add_argument('--verbose'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = """Leave it off please - hard to find a reasonable
                                amount of things that's neither too little nor
                                too much ... maybe the input for this should
                                actually be 'high 'low' or 'medium'?""")

parser.add_argument('--local'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Default: if on sherlock, don\'t run local')

parser.add_argument('--reset'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Nuke the DB first - needed if you make schema changes')

parser.add_argument('--add'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Try to add columns (if you make benign schema additions)')

parser.add_argument('--meta'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Make a meta.db')

parser.add_argument('--only'
                   ,default = ''
                   ,help    = 'Run only the (space separated) rules')

parser.add_argument('--xclude'
                   ,default = 'underlying ads_triple'
                   ,help    = 'Do not run the (space separated) rules')

parser.add_argument('--jobs'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Run all job_loading rules (takes time to rerun even if no new jobs)')

parser.add_argument('--catalog'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Run only the catalog job loading rule ')

parser.add_argument('--test'
                   ,default = ''
                   ,nargs   = 2
                   ,help    = 'Give a Tranform Name and an input dictionary (as parsable string) to test')

parser.add_argument('--parallel'
                   ,default = True
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Parallelize processing of multiple Query inputs. (Turn off if using pdb)')
################################################################################
def load_jobs(jobs     : bool = False
             ,catalog  : bool = True
             ,add      : bool = False
             ,reset    : bool = False
             ,parallel : bool = True
             ,only     : str  = ''
             ,xclude   : str  = ''
             ) -> None:
        """
        Calls the main function from within Python (bypass command line args)
        See parser description for input variable meanings
        """

        main({'jobs':jobs,'catalog':catalog,'add':add,'reset':reset
             ,'parallel':parallel,'only':only,'xclude':xclude})

def main(direct_args : dict = {})->None:
    """
    Executes DB modifying procedure described by a DBG instance
    """
    # Get args
    #------------
    parser.set_defaults(**direct_args)
    args = parser.parse_args() # get args from command line

    if direct_args:
        print(direct_args)
    print('args are: \n'+pformat(args))


    if args.test:
        # Don't interact with database
        #------------------------------------
        dic = literal_eval(args.test[1])
        assert isinstance(dic,dict), "We need a dictionary to test a Rule's input"
        dbg.test_input(args.test[0],dic)
    else:
        # Actually interact with database
        #------------------------------------
        if args.local:
            db = ConnectInfo()
        else:
            with open(environ['DB_JSON'],'r') as f:
                db = ConnectInfo(**load(f))

        if args.meta:
            dbg.test_meta() # create a meta-db

        # Determine which job-loading rules to include
        #------------------------------------------------------------
        if not args.jobs:
            args.xclude+=''' stray_gpaw_logfiles
                             stray_qe_logfiles
                             stray_chargemol_logfiles ''' # these T's ONLY run if --jobs

        if (not args.reset) and (not (args.catalog or args.jobs)):
            args.xclude+=''' catalog ''' # this run if EITHER --jobs OR --catalog...provided we didn't just nuke DB

        # Process string inputs
        #------------------------
        only   = set(args.only.split())
        xclude = set(args.xclude.split()) - only

        # Verify every  rule excluded/included by name actually exists
        # if first 5 chars match or within edit distance of 5, suggest alternatives
        #------------------------------------------------------------------------------------
        for w in (only | xclude):
            match = False
            close = []
            for n in [t.name for t in dbg.rules]:
                d = levenshteinDistance(w,n)
                upW,upN = max(len(w),5), max(len(n),5) # variables for safe indexing
                if d == 0:
                    match = True
                    break
                elif d < 5 or w[:upW] == n[:upN]:
                    close.append(n) # keep track of near-misses
            if not match:
                did_you = "Did you mean %s"%close if close else ''
                raise ValueError("No match found for %s\n%s"%(w,did_you))

        dbg.run_all(db
                   ,only    = only
                   ,xclude  = xclude
                   ,reset   = args.reset
                   ,add     = args.add
                   ,verbose = args.verbose
                   ,parallel= args.parallel)

if __name__ == '__main__':
    main()



from typing import Optional,List,Union,Tuple
# External Modules
import os
from os.path import join,exists
from ast import literal_eval
# Internal Modules
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.rule import Rule
from dbgen.support.datatypes.func      import Func
from dbgen.support.datatypes.dbg       import  DBG
from dbgen.support.datatypes.misc      import  Arg
from dbgen.support.datatypes.temp      import Temp,TempFunc,Const
from dbgen.support.datatypes.metatemp  import MetaTemp,Block
from dbgen.support.datatypes.sqltypes  import SQLType,Int,Varchar,Text,Decimal
from dbgen.support.utils               import sqlite_select
from dbgen.core.misc                   import maybe
###############################################################################


# Helper function
#----------------
def splt(x:Optional[str],y:Optional[str]=None)->List[Optional[str]]:
    if x is None:
        return [None]*100
    else:
        delim = y if y else ','
        return x.split(delim)  # type: ignore

def extract_dbg(pth:str)->DBG:
    """ Extract everything """
    if exists(pth):
        return DBG(tables = get_tables(pth)
                  ,rules  =  get_rules(pth)
                  ,views  = get_views(pth))
    else:
        raise IOError("No .db file found at "+pth)

def get_views(pth:str)->List[View]:
    q = "SELECT name,QUERY from views"
    qOut = sqlite_select(pth,q)
    return [View(n,q) for n,q in qOut]

def get_tables(pth:str)->List[Table]:
    colQ="""SELECT T.tab_id,T.name,T.desc
                    ,group_concat(C.name,'|')
                    ,group_concat(C.type,'|')
                    ,group_concat(C.nnull,'|')
                    ,group_concat(C.pk,'|')
                    ,group_concat(C.uniq,'|')
                    ,group_concat(C.auto,'|')
                    ,group_concat(C.ind,'|')
                    ,group_concat(COALESCE(C.virt,'None'),'|')
                    ,group_concat(COALESCE(C.dfault,'None'),'|')
            FROM  tab T
            LEFT JOIN col C USING (tab_id)
            GROUP BY T.tab_id
            """
    rawtabs = sqlite_select(pth,colQ)
    table_dict = {}

    for (t_id,tn,td,cn,ct,cnn,cpk,cu,ca,ci,cv,cd) in rawtabs:

        # Add columns
        #-----------
        cols     = []
        spltCols = [cn,ct,cnn,cpk,cu,ca,ci,cv,cd]
        splttr   = lambda x: splt(x,'|')
        for n,t,nn,pk,u,a,i,v,d in zip(*map(splttr,spltCols)):
            typ = SQLType.from_str(t)
            if isinstance(typ,Int):
                d = None if d=='None' else int(d)
            elif isinstance(typ,Decimal):
                d = None if d=='None' else float(d)
            elif isinstance(typ,(Text,Varchar)):
                pass
            else:
                raise NotImplementedError("New type? %s %s"%(str(typ),typ.__class__))

            cols.append(Col(n,typ,nn=='1',pk=='1',u=='1',a=='1',i=='1'
                           ,None if v=="None" else v,None if d=="None" else d))

        # Add FKs
        #--------
        pkQ = """SELECT group_concat(FC.name),TT.name,group_concat(TC.name)
                from fk FK
                join fk_cols FKC USING (tab_id,fk_id)
                join col FC USING  (tab_id,col_id)
                join tab FT USING (tab_id)
                join col TC on FKC.to_tab = TC.tab_id and FKC.to_col = TC.col_id
                join tab TT on FKC.to_tab = TT.tab_id
                where FKC.tab_id = ?
                group by FK.fk_id
                order by FK.fk_id"""
        pkOut = sqlite_select(pth,pkQ,[t_id])
        fks = []
        for fcnames,ttname,tcnames in pkOut:
            fks.append(FK(splt(fcnames),ttname,splt(tcnames))) # type: ignore

        table_dict[tn] = Table(tn,td,cols,fks)

    return list(table_dict.values())

def get_rules(pth:str)->List[Rule]:
    """
    Getting rules...
    """
    #print('\tGetting rules...')

    q = """SELECT rule_id,name,query FROM rule """
    qOut = sqlite_select(pth,q)
    output      = [] # type: List[Rule]
    # For each rule, extract its template, store in tempDict
    for (t_id,name,query) in qOut:
        t = Rule(name     = name
                     ,query     = query
                     ,metatemp  = get_metatemp(pth,t_id)
                     ,template  = get_temp(pth,t_id))
        output.append(t)
    return output

def get_metatemp(pth    : str
                ,t_id   : int
                ) -> MetaTemp:
        """
        Parses a metatemp from meta.db
        """
        # Get Blocks
        #--------------
        #print('\t\tGetting metatemp...')
        blockQ = """SELECT B.name,B.text
                        ,GROUP_CONCAT(BA.name)
                        ,GROUP_CONCAT(COALESCE(BA.ind,'None'))
                        FROM block B
                        LEFT JOIN blockarg BA USING (rule_id,block_id)
                        WHERE B.rule_id = ?
                        GROUP BY B.block_id
                        ORDER BY B.block_id """

        blocksOut = sqlite_select(pth,blockQ,[t_id])
        blocks = [] # type: List[Block]
        for name,txt,names,inds in blocksOut:
            if names is None:
                args = [] # type: List[Arg]
            else:
                args = [Arg(n,None if i=='None' else int(i)) for n,i in zip(*map(splt,[names,inds]))]
            blocks.append(Block(name,txt,args))

        return MetaTemp(blocks)

def get_temp(pth    : str
            ,t_id   : int
            ) -> Temp:
    """
    Parses a template from meta.db
    """
    #print('\t\tGetting temp...')

    # Get TempFuncs
    #--------------
    tfQ = """SELECT F.name,TF.name
                    ,GROUP_CONCAT(TFA.name)
                    ,GROUP_CONCAT(COALESCE(TFA.ind,'None'))
                    FROM tempfunc TF

                    JOIN func F USING (func_id)

                    LEFT JOIN tempfuncarg TFA USING  (rule_id,tempfunc_id)
                    WHERE TF.rule_id = ?
                    GROUP BY TF.tempfunc_id """

    tfOut = sqlite_select(pth,tfQ,[t_id])

    tempfuncs = [] # type: List[TempFunc]
    stdout    = [] # type: ignore

    for (fName,name,tfa_names,tfa_inds) in tfOut:
        if tfa_names == None:
            args = [] # type: List[Arg]
        else:
            args = [Arg(n,None if i=='None' else int(i)) for n,i in zip(*map(splt,[tfa_names,tfa_inds]))]
        tempfuncs.append(TempFunc(fName,name,args))

    # Get Constants
    #--------------

    constQ = """SELECT name,datatype,val
                FROM constants
                WHERE rule_id = ?"""
    constOut = sqlite_select(pth,constQ,[t_id])

    constants = [] # type: List[Tuple[str,Const]]
    for name,dt,val in constOut:
        if 'float' in dt:
            val = float(val)
        elif 'int' in dt:
            val = int(val)
        elif 'dict' in dt:
            val = literal_eval(val)
        elif 'str' in dt:
            pass
        elif 'callable' in dt:
            raise NotImplementedError
        else:
            raise ValueError('bad datatype? '+dt)
        constants.append((name,Const(val)))

    return Temp(tempfuncs=tempfuncs,constants=dict(constants))



import setuptools, re # type: ignore

with open('dbgen/__init__.py') as fd:
    version = re.search("__version__ = '(.*)'", fd.read()).group(1)

setuptools.setup(
    name="CC-dbgen"
    ,version=version
    ,url="https://github.com/ksb/dbgen"
    ,author="Kris Brown"
    ,author_email="ksb@stanford.edu"
    ,description="Tools for computational catalyst research"
    ,license='APACHE LICENSE, VERSION 2.0'
    ,packages=['dbgen'
              ,'dbgen.core'
              ,'dbgen.data'
              ,'dbgen.inputs'
              ,'dbgen.new_inputs'
              ,'dbgen.scripts'
              ,'dbgen.support'
    ]
    ,include_package_data=True
    # ,entry_points={'console_scripts': ['catalog=catalog.cli.main:main']}
    ,install_requires=['ase'
    ,'sqlparse'
    ,'networkx'
    ,'python-sql'
    ,'mysqlclient'
    ,'pymatgen'
    ,'tqdm'
    ,'PyOpenGL'
    ,'PyQt5'
    ]
    ,python_requires='>3.6, <4'
    ,classifiers=[
        'Development Status :: 4 - Beta'
       ,'Intended Audience :: Developers'
       ,'Topic :: Scientific/Engineering :: Chemistry'
       ,'Programming Language :: Python :: 3.6'
    ]
)



# Internal Modules
from dbgen.support.datatypes.sqltypes     import Varchar,Date
from dbgen.support.datatypes.select       import Select
from dbgen.support.datatypes.temp         import SimpleTemp,SimplePipe,Const
from dbgen.support.datatypes.object       import Object,Attr,Universe
from dbgen.support.datatypes.relation     import Relation
from dbgen.support.datatypes.insertupdate import Insert
from dbgen.support.datatypes.constraint   import NULL,IN,HAS
from dbgen.support.datatypes.temp         import Temp,TempFunc
from dbgen.support.datatypes.misc         import Arg,noIndex

from dbgen.new_inputs.objects             import *
from dbgen.core.lists                     import flatten

from dbgen.scripts.IO.get_catalog_logfiles      import get_catalog_logfiles
from dbgen.scripts.IO.anytraj                   import anytraj
from dbgen.scripts.IO.parse_mendeleev           import parse_mendeleev
from dbgen.scripts.IO.get_job_metadata          import get_job_metadata
from dbgen.scripts.Pure.Misc.dftcode_dependent  import dftcode_dependent
from dbgen.scripts.Pure.Misc.identity           import identity
from dbgen.scripts.Pure.Atoms.cell_info         import cell_info
from dbgen.scripts.Pure.Atoms.json_to_traj      import json_to_traj
from dbgen.scripts.Pure.Atoms.get_system_type   import get_system_type
from dbgen.scripts.Pure.Load.populate_storage   import populate_storage
from dbgen.scripts.Pure.Load.parse_pw_gpaw      import parse_pw_gpaw
from dbgen.scripts.Pure.Load.parse_pw_qe        import parse_pw_qe
from dbgen.scripts.Pure.Load.parse_pw_vasp      import parse_pw_vasp
from dbgen.scripts.Pure.Load.parse_xc_gpaw      import parse_xc_gpaw
from dbgen.scripts.Pure.Load.parse_xc_qe        import parse_xc_qe
from dbgen.scripts.Pure.Load.parse_xc_vasp      import parse_xc_vasp
from dbgen.scripts.Pure.Load.parse_psp_gpaw     import parse_psp_gpaw
from dbgen.scripts.Pure.Load.parse_psp_qe       import parse_psp_qe
from dbgen.scripts.Pure.Load.parse_psp_vasp     import parse_psp_vasp
from dbgen.scripts.Pure.Load.normalize_xc       import normalize_xc
from dbgen.scripts.Pure.Load.normalize_psp      import normalize_psp

##############################################################################

# Constants
#----------
pop_job_cols = ['stordir','logfile','code','log','pwinp'
               ,'potcar','poscar','kptcar','paramdict']

mdata_input = ['stordir','code','paramdict']
mdata_cols  = ['user','timestamp','working_directory'
              ,'job_type','job_name']

elem_cols  =  [ 'element_id', 'symbol', 'name', 'atomic_weight','atomic_radius'
              , 'phase','evaporation_heat', 'pointgroup','spacegroup'
              , 'melting_point', 'metallic_radius', 'vdw_radius'
              , 'density', 'en_allen', 'is_radioactive'
              , 'lattice_struct' , 'fusion_heat'
              , 'econf', 'period', 'covalent_radius_bragg'
              , 'geochemical_class', 'abundance_crust', 'heat_of_formation'
              , 'electron_affinity', 'atomic_volume',  'boiling_point'
              , 'proton_affinity', 'covalent_radius_slater'
              , 'lattice_constant', 'dipole_polarizability'
              , 'en_ghosh', 'thermal_conductivity', 'group_id', 'en_pauling'
              , 'gas_basicity','abundance_sea']

cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']


dfts       = ['gpaw','qe','vasp']
parsed     = ['pw',  'xc','psp']

calc_funcs = [parse_pw_gpaw,parse_pw_qe,parse_pw_vasp # type: ignore
             ,parse_xc_gpaw,parse_xc_qe,parse_xc_vasp
             ,parse_psp_gpaw,parse_psp_qe,parse_psp_vasp
             ,normalize_xc,normalize_psp,identity]

##############################################################################

relax_calc_func = Temp(constants = {f.__name__:Const(f) for f in calc_funcs}
    ,tempfuncs = [TempFunc(func = dftcode_dependent
                          ,name = x
                          ,args = noIndex(['dftcode'
                                          ,'log'
                                          ,'parse_%s_gpaw'%x
                                          ,'parse_%s_qe'%x
                                          ,'parse_%s_vasp'%x
                                          ,'identity'
                                          ,'identity' if x =='pw' else 'normalize_%s'%x]))
                    for x in parsed])

########################################################################
loading_relations = [
    Relation(name    = 'catalog'
            ,select  = uni.select(attrs       = [roots.root]                 # type: ignore
                             ,constraints = [roots.label == "'catalog'"  # type: ignore
                                            ,roots.active ==1 ])         # type: ignore
            ,func    = SimpleTemp(func    = get_catalog_logfiles
                                 ,inputs  = ['root']
                                 ,outputs = pop_job_cols)
            ,inserts = [alljob.insert(schema,pop_job_cols)])

   ,Relation(name    = 'populate_storage'
            ,func    = SimpleTemp(func    = populate_storage
                                 ,outputs = ['root','code','label'])
            ,inserts = [roots.insert(schema,['root','code','label'])])

   ,Relation(name    = 'job_metadata'
            ,select  = Select(schema
                             ,attrs       = alljob.select(schema,mdata_input)
                             ,constraints = [NULL(alljob.user)     # type: ignore
                                            ,alljob.deleted==0])   # type: ignore
            ,func    = SimpleTemp(func    = get_job_metadata
                                 ,inputs  = mdata_input
                                 ,outputs = mdata_cols)
            ,inserts = [alljob.update(schema,mdata_cols)])

   ,Relation(name   = 'anytraj'
            ,select = Select(schema
                             ,attrs       = alljob.select(schema,['stordir'])
                             ,constraints = [NULL(alljob.anytraj)  # type: ignore
                                            ,alljob.deleted==0])   # type: ignore
            ,func   = SimpleTemp(anytraj,['stordir'],['anytraj'])
            ,inserts= [alljob.update(schema,['anytraj'])])

    ,Relation(name = 'mendeleev'
             ,func = SimpleTemp(parse_mendeleev,outputs = elem_cols)
             ,inserts = [element.insert(schema,elem_cols)])

    ,Relation(name ='cell_info'
             ,select = Select(schema
                             ,attrs = cell.select(schema,cell_cols))
             ,func = SimpleTemp(cell_info,cell_cols,['surface_area','volume'])
             ,inserts = [cell.insert(schema,['surface_area','volume'])])

   ,Relation(name    = 'relax_calc'
            ,select  = Select(schema
                             ,attrs = alljob.select(schema,['log','code'])
                             ,constraints = [alljob.deleted == 0       # type: ignore
                                            ,HASNT(relax_job,calc)]) # type: ignore
            ,func    = relax_calc_func
            ,inserts = [calc.insert(schema,['dftcode','pw','xc','psp'])
                       ,relax_job.update_component(schema,calc)])

   ,Relation(name = 'systype'
            ,select = Select(schema
                            ,attrs       = [struct.raw]                 # type: ignore
                            ,constraints = [NULL(struct.system_type)])  # type: ignore
            ,func = SimplePipe([json_to_traj,get_system_type]
                               ,inputs  = ['raw']
                               ,outputs = ['system_type'])
            ,inserts = [struct.update(schema,['system_type'])])

   ,Relation(name    = 'relax_job'
            ,select  = Select(schema
                             ,attrs       = alljob.select(schema)
                             ,constraints = [IN(alljob.job_type   # type: ignore
                                            ,["'relax'","'latticeopt'","'vcrelax'"])])
            ,inserts = [relax_job.insert(schema)])

]

# if any ionic steps are in the traj table, then we assume all 4 tables have been
# taken care of
# populate_traj = Rule(name='populate_struct_traj_atom_cell'
#     ,query = """SELECT R.job_id,J.code AS dftcode,J.log,J.anytraj,J.poscar,J.stordir
#                  FROM relax_job AS R JOIN alljob AS J USING (job_id)
#                  WHERE R.job_id NOT IN (SELECT traj.job_id FROM traj)
#                  AND J.code != 'vasp' # TEMPORARY THING
#                  AND NOT J.deleted"""
#
#     ,template = Temp(constants = {'get_cell'              : Const('get_cell')
#                                    ,'get_atom'              : Const('get_atom')
#                                    ,'traj_to_json'          : Const('traj_to_json')
#                                    ,'get_traj_gpaw%s'%laz   : Const('get_traj_gpaw%s'%laz)
#                                    ,'get_traj_qe%s'%laz     : Const('get_traj_qe%s'%laz)
#                                    ,'get_traj_vasp%s'%laz   : Const('get_traj_vasp%s'%laz)
#                                    ,'identity'              : Const('identity')
#                                    ,'unzip'                 : Const('unzip')}
#         ,tempfuncs=[
#               TempFunc(triple'
#                       ,'anytraj_log_poscar'
#                       ,args = noIndex(['anytraj','log','poscar']))
#
#
#               ,TempFunc(dftcode_dependent'
#                        ,name = 'get_traj'
#                        ,args = noIndex(['dftcode'
#                                   ,'anytraj_log_poscar'
#                                   ,'get_traj_gpaw%s'%laz
#                                   ,'get_traj_qe%s'%laz
#                                   ,'get_traj_vasp%s'%laz
#                                   ,'identity'
#                                   ,'unzip'])) # returns ([ASEATOMS]m,[STEPNUM]m,[ENERGY]m,[TRAJATOMLIST]m)
#                ,TempFunc(map'
#                         ,name = 'cell'
#                         ,args = noIndex(['get_cell','structs_few'])) # returns ([ax]m,[ay]m,...)
#                ,TempFunc(map'
#                         ,name = 'structs_few'
#                         ,args = [Arg('traj_to_json')
#                                 ,Arg('get_traj', 0)]) # returns [raw_json]m
#                ,TempFunc(concat_map'
#                         ,name = 'atoms'
#                         ,args = noIndex(['get_atom','structs_few'])) # returns ([ASEATOMS]n,[Ind]n,[ELEMENT]n,[X]n,[Y]n,...)
#                ,TempFunc(identity'
#                         ,name = 'structs_many'
#                         ,args = [Arg('atoms', 0)])# returns [raw_json]n
#                 ,TempFunc(flatten'
#                          ,name = 'flattened_trajatoms'
#                          ,args = [Arg('get_traj',3)])  # [STEPNUM,Ind,FX,FY,FZ]n
#                 ,TempFunc(unzip'
#                          ,'trajatoms'
#                          ,args=[Arg('flattened_trajatoms')])]) # ([STEPNUM]n,[Ind]n,[FX]n,[FY]n,[FZ]n)
#
#     ,metatemp = MetaTemp([
#         Block(name = 'insert_cell'
#              ,text = mkInsCmd('cell',cell_cols,sqlite=False)
#              ,args = cell_args)
#
#        ,Block(name = 'query_cell'
#              ,text = ('SELECT C.cell_id FROM cell AS C WHERE '
#                      +addQs(['C.%s'%x for x in cell_cols],' AND '))
#              ,args = cell_args)
#
#        ,Block(name = 'insert_struct'
#              ,text = """INSERT INTO struct (raw,rawhash,cell_id)
#                            VALUES (%s,SHA2(%s,512),%s)
#                         ON DUPLICATE KEY UPDATE cell_id=cell_id"""
#              ,args = [Arg('structs_few')
#                       ,Arg('structs_few')
#                      ,Arg('query_cell',0)])
#
#        ,Block(name = 'query_struct_few'
#              ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
#              ,args = [Arg('structs_few')])
#
#        ,Block(name = 'query_struct_many'
#              ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
#              ,args = [Arg('structs_many')])
#
#        ,Block(name = 'insert_atom'
#              ,text = mkInsCmd('atom',['struct_id','atom_id','element_id'
#                                   ,'x','y','z','constrained','magmom','tag']
#                             ,sqlite=False)
#              ,args = [Arg('query_struct_many')]
#                     +[Arg('atoms',i) for i in range(1,9)])
#
#      ,Block(name = 'insert_traj'
#            ,text = mkInsCmd('traj',['job_id','traj_id','struct_id','energy','final'],sqlite=False)
#            ,args = [Arg('job_id')
#                    ,Arg('get_traj',1)
#                    ,Arg('query_struct_few')
#                    ,Arg('get_traj',2)
#                    ,Arg('get_traj',4)])
#
#     ,Block(name = 'insert_trajatom'
#           ,text = mkInsCmd('trajatom',['job_id','traj_id','struct_id','atom_id','fx','fy','fz'],sqlite=False)
#           ,args = [Arg('job_id')
#                   ,Arg('trajatoms',0)
#                   ,Arg('query_struct_many')]
#                  +[Arg('trajatoms',i) for i in range(1,5)])]))

relations = loading_relations



from dbgen.support.datatypes.table      import Table,Col,FK
from dbgen.support.datatypes.rule  import Rule
from dbgen.support.datatypes.misc       import Arg,noIndex
from dbgen.support.datatypes.metatemp   import Block,MetaTemp,SimpleUpdate,PureSQL
from dbgen.support.datatypes.temp       import Temp
from dbgen.support.datatypes.sqltypes   import Varchar,Decimal
from dbgen.support.utils                import mkInsCmd

##########################################################################################

stoich = Table('stoich'
        ,desc = 'Mapping table between jobs and reactions'
        ,cols = [Col('job_id',    pk=True)
                ,Col('rxn_id',    pk=True)
                ,Col('stoich',    nn=True)]
        ,fks = [FK('job_id','relax_job')
               ,FK('rxn_id','rxn')])

rxn = Table('rxn'
        ,desc = 'Any reaction'
        ,cols = [Col('rxn_id',                pk=True,auto=True)
                ,Col('name',        Varchar(),    nn=True,uniq=True)
                ,Col('is_adsorption',        nn=True)
                ,Col('dE',            Decimal())])

ads_triple = Table('ads_triple'
        ,desc='Triples of bare + complex + adsorbate'
        ,cols = [Col('bare_id',                pk=True)
                ,Col('complex_id',             pk=True)
                ,Col('adsorbate_id',           pk=True)
                ,Col('delta_e_surf',Decimal(),   nn=True)
                ,Col('calc_id',                nn=True)
                ,Col('metal_comp',Varchar(),       nn=True)
                ]
        ,fks = [FK('bare_id',       'relax_job',    'job_id')
               ,FK('complex_id',    'relax_job',    'job_id')
               ,FK('adsorbate_id',  'adsorbate',    'adsorbate_id')
               ,FK('calc_id',       'calc',         'calc_id')
               ])
tables = [stoich,rxn,ads_triple]
##########################################################################################
##########################################################################################
##########################################################################################
rxn_energies = Rule('rxn_energies'
    ,query = """SELECT S.rxn_id,SUM(T.energy * S.stoich) AS dE
                    FROM stoich    AS S
                    JOIN rxn         AS R USING (rxn_id)
                    JOIN finaltraj AS T USING (job_id)
                    WHERE R.dE IS NULL
                    GROUP BY S.rxn_id """
    ,metatemp = SimpleUpdate(rxn,['dE']))

################################################
# Pairs of surfaces that differ by an adsorbate
#################################################
pop_ads_triple=Rule('ads_triple'
    ,metatemp=PureSQL(["""
            INSERT INTO ads_triple (bare_id
                                   ,complex_id
                                   ,adsorbate_id
                                   ,delta_e_surf
                                   ,calc_id
                                   ,metal_comp)
                SELECT X.bare_id
                      ,X.complex_id
                      ,A.adsorbate_id
                      ,X.delta_e_surf
                      ,X.calc_id
                      ,X.metal_comp
                FROM
                (SELECT
                        F1.job_id                    AS complex_id
                        ,F2.job_id               AS bare_id
                        ,MIN(F1.energy - F2.energy)  AS delta_e_surf -- MIN actually doesn't matter - schema doesn't know Finaltraj as 1-1 relationship
                        ,J1.calc_id                  AS calc_id
                        ,MIN(S1.metal_comp)          AS metal_comp -- MIN actually doesn't matter
                        , GROUP_CONCAT(CONCAT('[',C.element_id,
                                              ',',C.count - COALESCE(C2.count, 0),
                                              ']')
                                    ORDER BY C.element_id ASC) AS composition
                        FROM
                            finaltraj F1
                                JOIN struct        AS S1  ON F1.struct_id=S1.struct_id
                                JOIN surface       AS SS1 ON S1.struct_id=SS1.struct_id
                                JOIN relax_job     AS J1  ON F1.job_id=J1.job_id
                                JOIN struct        AS S2  USING (metal_comp)
                                JOIN finaltraj     AS F2  ON F2.struct_id=S2.struct_id
                                JOIN relax_job     AS J2  ON F2.job_id=J2.job_id AND J2.calc_id=J1.calc_id
                                JOIN surface       AS SS2 ON S2.struct_id=SS2.struct_id

                                JOIN      struct_composition AS C   ON (C.struct_id=S1.struct_id)
                                LEFT JOIN struct_composition AS C2  ON C2.struct_id = S2.struct_id AND C2.element_id=C.element_id
                        WHERE
                                -- (F1.job_id,F2.job_id) NOT IN (SELECT complex_id,bare_id FROM ads_triple) AND
                                S1.n_atoms > S2.n_atoms
                                AND S1.n_elems >= S2.n_elems
                                AND C.element_id IN (1 , 2, 6, 7, 8)                 -- already know they have same metal stoich -
                                AND C.count - COALESCE(C2.count, 0) != 0 -- ignore in case the have the same # of some nonmetal
                        GROUP BY complex_id,bare_id) AS X
                 JOIN adsorbate  AS A ON A.composition = X.composition

            ON DUPLICATE KEY UPDATE bare_id=X.bare_id"""]))


rules = [rxn_energies,pop_ads_triple]



import numpy as np 
qubic_solids_26 = ['Li-bcc','Na-bcc','K-bcc','Rb-bcc',
                   'Ca-fcc','Sr-fcc','Ba-bcc',
                   'Nb-bcc','Ta-bcc',
                   'Mo-bcc','W-bcc','Fe-bcc',
                   'Rh-fcc','Ir-fcc',
                   'Ni-fcc','Pd-fcc','Pt-fcc',
                   'Cu-fcc','Ag-fcc','Au-fcc',
                   'Al-fcc','Pb-fcc',              #KRIS: WHY WERE Pb-fcc and Au-fcc commented out???
                   'C-diamond','Si-diamond','Ge-diamond','Sn-diamond']
qubic_solids_27 = np.append(qubic_solids_26, 'V-bcc')

hcp_solids_10 = ['Cd-hcp', 'Co-hcp', 'Os-hcp', 'Ru-hcp', 'Zn-hcp',
                 'Zr-hcp', 'Sc-hcp', 'Be-hcp', 'Mg-hcp', 'Ti-hcp']


# ZPAE (semi-empirical) corrected lattice parameters from Haas, Tran, Blaha, PRB 79, 085104 (2009)
# and ZPAE (phonon) corrected lattice parameters + Debye temps + bulk moduli (GPa) from Hao et al., PRB 85, 014111 (2012)
# cohesive energies (at 0 K) and Debye temperatures from Kittel and Hao et al.
data = {
# Cubic pure solids
'Li-bcc': {
    'name': 'Li-bcc',
    'lattice parameter': 3.451,
    'pbe-lp': 3.429,
    'cohesive energy': 1.63,
    'debye temperature': 344.,
    'bulk modulus': 13.,
    'symbols': 'Li',
    'latex name': "Li",
    'structure':'bcc',
    'magmom': None},
'Na-bcc': {
    'name': 'Na-bcc',
    'lattice parameter': 4.207,
    'pbe-lp': 4.197,
    'cohesive energy': 1.113,
    'bulk modulus': 7.5,
    'debye temperature': 158.,
    'symbols': 'Na',
    'latex name': "Na",
    'structure':'bcc',
    'magmom': None},
'K-bcc': {
    'name': 'K-bcc',
    'lattice parameter': 5.211,
    'pbe-lp': 5.281,
    'cohesive energy': 0.934,
    'bulk modulus': 3.7,
    'debye temperature': 91.,
    'symbols': 'K',
    'latex name': "K",
    'structure':'bcc',
    'magmom': None},
'Rb-bcc': {
    'name': 'Rb-bcc',
    'lattice parameter': 5.580,
    'pbe-lp': 5.665,
    'cohesive energy': 0.852,
    'bulk modulus': 3.1,
    'debye temperature': 56.,
    'symbols': 'Rb',
    'latex name': "Rb",
    'structure':'bcc',
    'magmom': None},
'Ca-fcc': {
    'name': 'Ca-fcc',
    'lattice parameter': 5.555,
    'pbe-lp': 5.521,
    'cohesive energy': 1.84,
    'bulk modulus': 18.36,
    'debye temperature': 230.,
    'symbols': 'Ca',
    'latex name': "Ca",
    'structure':'fcc',
    'magmom': None},
'Sr-fcc': {
    'name': 'Sr-fcc',
    'lattice parameter': 6.042,
    'pbe-lp': 6.013,
    'cohesive energy': 1.72,
    'bulk modulus': 12.36,
    'debye temperature': 147.,
    'symbols': 'Sr',
    'latex name': "Sr",
    'structure':'fcc',
    'magmom': None},
'Ba-bcc': {
    'name': 'Ba-bcc',
    'lattice parameter': 5.004,
    'pbe-lp': 5.022,
    'cohesive energy': 1.90,
    'bulk modulus': 9.3,
    'debye temperature': 110.,
    'symbols': 'Ba',
    'latex name': "Ba",
    'structure':'bcc',
    'magmom': None},
'V-bcc': {
    'name': 'V-bcc',
    'lattice parameter': 3.024,
    'pbe-lp': 2.997,
    'cohesive energy': 5.31,
    'bulk modulus': 161.9,
    'debye temperature': 380.,
    'symbols': 'V',
    'latex name': "V",
    'structure':'bcc',
    'magmom': None},
'Nb-bcc': {
    'name': 'Nb-bcc',
    'lattice parameter': 3.293,
    'pbe-lp': 3.310,
    'cohesive energy': 7.57,
    'bulk modulus': 170.2,
    'debye temperature': 275.,
    'symbols': 'Nb',
    'latex name': "Nb",
    'structure':'bcc',
    'magmom': None},
'Ta-bcc': {
    'name': 'Ta-bcc',
    'lattice parameter': 3.299,
    'pbe-lp': 3.347,
    'cohesive energy': 8.10,
    'bulk modulus': 194.,
    'debye temperature': 240.,
    'symbols': 'Ta',
    'latex name': "Ta",
    'structure':'bcc',
    'magmom': None},
'Mo-bcc': {
    'name': 'Mo-bcc',
    'lattice parameter': 3.141,
    'pbe-lp': 3.161,
    'cohesive energy': 6.82,
    'bulk modulus': 272.5,
    'debye temperature': 450.,
    'symbols': 'Mo',
    'latex name': "Mo",
    'structure':'bcc',
    'magmom': None},
'W-bcc': {
    'name': 'W-bcc',
    'lattice parameter': 3.161,
    'pbe-lp': 3.170,
    'cohesive energy': 8.90,
    'bulk modulus': 314.,
    'debye temperature': 400.,
    'symbols': 'W',
    'latex name': "W",
    'structure':'bcc',
    'magmom': None},
'Fe-bcc': {
    'name': 'Fe-bcc',
    'lattice parameter': 2.855,
    'pbe-lp': 2.834,
    'cohesive energy': 4.28,
    'bulk modulus': 168.3,
    'debye temperature': 470.,
    'symbols': 'Fe',
    'latex name': "Fe",
    'structure':'bcc',
    'magmom': 2.22},
'Rh-fcc': {
    'name': 'Rh-fcc',
    'lattice parameter': 3.793,
    'pbe-lp': 3.827,
    'cohesive energy': 5.75,
    'bulk modulus': 269.,
    'debye temperature': 480.,
    'symbols': 'Rh',
    'latex name': "Rh",
    'structure':'fcc',
    'magmom': None},
'Ir-fcc': {
    'name': 'Ir-fcc',
    'lattice parameter': 3.832,
    'pbe-lp': 3.872,
    'cohesive energy': 6.94,
    'bulk modulus': 355.,
    'debye temperature': 420.,
    'symbols': 'Ir',
    'latex name': "Ir",
    'structure':'fcc',
    'magmom': None},
'Ni-fcc': {
    'name': 'Ni-fcc',
    'lattice parameter': 3.509,
    'pbe-lp': 3.517,
    'cohesive energy': 4.44,
    'bulk modulus': 188.,
    'debye temperature': 450.,
    'symbols': 'Ni',
    'latex name': "Ni",
    'structure':'fcc',
    'magmom': 0.64},
'Pd-fcc': {
    'name': 'Pd-fcc',
    'lattice parameter': 3.876,
    'pbe-lp': 3.932,
    'cohesive energy': 3.89,
    'bulk modulus': 195.,
    'debye temperature': 274.,
    'symbols': 'Pd',
    'latex name': "Pd",
    'structure':'fcc',
    'magmom': None},
'Pt-fcc': {
    'name': 'Pt-fcc',
    'lattice parameter': 3.913,
    'pbe-lp': 3.971,
    'cohesive energy': 5.84,
    'bulk modulus': 278.3,
    'debye temperature': 240.,
    'symbols': 'Pt',
    'latex name': "Pt",
    'structure':'fcc',
    'magmom': None},
'Cu-fcc': {
    'name': 'Cu-fcc',
    'lattice parameter': 3.595,
    'pbe-lp': 3.630,
    'cohesive energy': 3.49,
    'bulk modulus': 142.,
    'debye temperature': 343.,
    'symbols': 'Cu',
    'latex name': "Cu",
    'structure':'fcc',
    'magmom': None},
'Ag-fcc': {
    'name': 'Ag-fcc',
    'lattice parameter': 4.063,
    'pbe-lp': 4.150,
    'cohesive energy': 2.95,
    'bulk modulus': 109.,
    'debye temperature': 225.,
    'symbols': 'Ag',
    'latex name': "Ag",
    'structure':'fcc',
    'magmom': None},
'Au-fcc': {
    'name': 'Au-fcc',
    'lattice parameter': 4.061,
    'pbe-lp': 4.147,
    'cohesive energy': 3.81,
    'bulk modulus': 180.,
    'debye temperature': 165.,
    'symbols': 'Au',
    'latex name': "Au",
    'structure':'fcc',
    'magmom': None},
'Al-fcc': {
    'name': 'Al-fcc',
    'lattice parameter': 4.019,
    'pbe-lp': 4.037,
    'cohesive energy': 3.39,
    'bulk modulus': 79.4,
    'debye temperature': 428.,
    'symbols': 'Al',
    'latex name': "Al",
    'structure':'fcc',
    'magmom': None},
'C-diamond': {
    'name': 'C-diamond',
    'lattice parameter': 3.555,
    'pbe-lp': 3.571,
    'cohesive energy': 7.37,
    'bulk modulus': 443.,
    'debye temperature': 2230.,
    'symbols': 'C',
    'latex name': "C (diamond)",
    'structure':'diamond',
    'magmom': None},
'Si-diamond': {
    'name': 'Si-diamond',
    'lattice parameter': 5.422,
    'pbe-lp': 5.468,
    'cohesive energy': 4.63,
    'bulk modulus': 99.2,
    'debye temperature': 645.,
    'symbols': 'Si',
    'latex name': "Si (diamond)",
    'structure':'diamond',
    'magmom': None},
'Ge-diamond': {
    'name': 'Ge-diamond',
    'lattice parameter': 5.644,
    'pbe-lp': 5.764,
    'cohesive energy': 3.85,
    'bulk modulus': 75.8,
    'debye temperature': 374.,
    'latex name': "Ge",
    'symbols': 'Ge',
    'structure':'diamond',
    'magmom': None},
'Sn-diamond': {
    'name': 'Sn-fcc',
    'lattice parameter': 6.476,
    'pbe-lp': 6.659,
    'cohesive energy': 3.14,
    'bulk modulus': 53.1,
    'debye temperature': 200.,
    'symbols': 'Sn',
    'latex name': "Sn",
    'structure':'diamond',
    'magmom': None},
'Pb-fcc': {
    'name': 'Pb-fcc',
    'lattice parameter': 4.912,
    'pbe-lp': 5.040,
    'cohesive energy': 2.03,
    'bulk modulus': 48.8,
    'debye temperature': 105.,
    'symbols': 'Pb',
    'latex name': "Pb",
    'structure':'fcc',
    'magmom': None},
# hcp pure solids
'Cd-hcp': {
    'name': "Cd-hcp",
    'lattice parameter': 2.979,
    'lattice parameter 2': 5.620,
    'revtpss-lp': 2.95057912237,
    'cohesive energy': 1.16,
    'bulk modulus kittel': 46.7, # GPa
    'symbols': 'Cd',
    'latex name': "Cd",
    'structure':'hcp',
    'magmom': None},
'Co-hcp': {
    'name': "Co-hcp",
    'lattice parameter': 2.507,
    'lattice parameter 2': 4.069,
    'revtpss-lp':  2.45918897125,
    'cohesive energy': 4.39,
    'bulk modulus kittel': 191.4, # GPa
    'symbols': 'Co',
    'latex name': "Co",
    'structure':'hcp',
    'magmom': 1.72},
    #'gpaw-conv': [1.e-5, 1.e-6, 1.e-4]},
'Os-hcp': {
    'name': "Os-hcp",
    'lattice parameter': 2.734,
    'lattice parameter 2': 4.392,
    'revtpss-lp': 2.73044397786,
    'cohesive energy': 8.17,
    'bulk modulus kittel': 418., # GPa
    'symbols': 'Os',
    'latex name': "Os",
    'structure':'hcp',
    'magmom': None},
'Ru-hcp': {
    'name': "Ru-hcp",
    'lattice parameter': 2.706,
    'lattice parameter 2': 4.282,
    'revtpss-lp':  2.7191530774,
    'cohesive energy': 6.74,
    'bulk modulus kittel': 320.8, # GPa
    'symbols': 'Ru',
    'latex name': "Ru",
    'structure':'hcp',
    'magmom': None},
'Zn-hcp': {
    'name': "Zn-hcp",
    'lattice parameter': 2.665,
    'lattice parameter 2': 4.947,
    'revtpss-lp': 2.59872461224,
    'cohesive energy': 1.35,
    'bulk modulus kittel': 59.8, # GPa
    'symbols': 'Zn',
    'latex name': "Zn",
    'structure':'hcp',
    'magmom': None},
'Ti-hcp': {
    'name': "Ti-hcp",
    'lattice parameter': 2.951,
    'lattice parameter 2': 4.684,
    'revtpss-lp': 2.92208023815,
    'cohesive energy': 4.85,
    'bulk modulus kittel': 105.1, # GPa
    'symbols': 'Ti',
    'latex name': "Ti",
    'structure':'hcp',
    'magmom': None,
    'gpaw-conv': [1.e-6, 1.e-4, 1.e-6]},
'Zr-hcp': {
    'name': "Zr-hcp",
    'lattice parameter': 3.232,
    'lattice parameter 2': 5.148,
    'revtpss-lp': 3.23504460776,
    'cohesive energy': 6.25,
    'bulk modulus kittel': 83.3, # GPa
    'symbols': 'Zr',
    'latex name': "Zr",
    'structure':'hcp',
    'magmom': None},
'Sc-hcp': {
    'name': "Sc-hcp",
    'lattice parameter': 3.309,
    'lattice parameter 2': 5.268,
    'revtpss-lp': 3.30710158266,
    'cohesive energy': 3.90,
    'bulk modulus kittel': 43.5, # GPa
    'symbols': 'Sc',
    'latex name': "Sc",
    'structure':'hcp',
    'magmom': None},
'Be-hcp': {
    'name': "Be-hcp",
    'lattice parameter': 2.286,
    'lattice parameter 2': 3.585,
    'revtpss-lp': 2.27878331698,
    'cohesive energy': 3.32,
    'bulk modulus kittel': 100.3, # GPa
    'symbols': 'Be',
    'latex name': "Be",
    'structure':'hcp',
    'magmom': None},
'Mg-hcp': {
    'name': "Mg-hcp",
    'lattice parameter': 3.209,
    'lattice parameter 2': 5.211,
    'revtpss-lp': 3.17456515719,
    'cohesive energy': 1.51,
    'bulk modulus kittel': 35.4, # GPa
    'symbols': 'Mg',
    'latex name': "Mg",
    'structure':'hcp',
    'magmom': None},
# compounds
'LiH-b1': {
    'name': 'LiH-b1',
    'lattice parameter': 3.979,
    'pbe-lp': 4.006,
    'cohesive energy': 2.49,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'LiH',
    'latex name': "LiH",
    'structure':'rocksalt',
    'magmom': None},
'LiF-b1': {
    'name': 'LiF-b1',
    'lattice parameter': 3.974,
    'pbe-lp': 4.064,
    'cohesive energy': 4.46,
    'debye temperature': 732.,
    'bulk modulus': 69.8,
    'symbols': 'LiF',
    'latex name': "LiF",
    'structure':'rocksalt',
    'magmom': None},
'LiCl-b1': {
    'name': 'LiCl-b1',
    'lattice parameter': 5.072,
    'pbe-lp': 5.147,
    'cohesive energy': 3.59,
    'debye temperature': 422.,
    'bulk modulus': 35.4,
    'symbols': 'LiCl',
    'latex name': "LiCl",
    'structure':'rocksalt',
    'magmom': None},
'NaF-b1': {
    'name': 'NaF-b1',
    'lattice parameter': 4.570,
    'pbe-lp': 4.700,
    'cohesive energy': 3.97,
    'debye temperature': 492.,
    'bulk modulus': 51.4,
    'symbols': 'NaF',
    'latex name': "NaF",
    'structure':'rocksalt',
    'magmom': None},
'NaCl-b1': {
    'name': 'NaCl-b1',
    'lattice parameter': 5.565,
    'pbe-lp': 5.695,
    'cohesive energy': 3.34,
    'debye temperature': 321.,
    'bulk modulus': 26.6,
    'symbols': 'NaCl',
    'latex name': "NaCl",
    'structure':'rocksalt',
    'magmom': None},
'MgO-b1': {
    'name': 'MgO-b1',
    'lattice parameter': 4.188,
    'pbe-lp': 4.255,
    'cohesive energy': 5.20,
    'debye temperature': 946.,
    'bulk modulus': 165.,
    'symbols': 'MgO',
    'latex name': "MgO",
    'structure':'rocksalt',
    'magmom': None},
'MgS-b1': {
    'name': 'MgS-b1',
    'lattice parameter': 5.188,
    'pbe-lp': 5.228,
    'cohesive energy': 4.07,
    'debye temperature': 650.,
    'bulk modulus': 80.,
    'symbols': 'MgS',
    'latex name': "MgS",
    'structure':'rocksalt',
    'magmom': None},
'CaO-b1': {
    'name': 'CaO-b1',
    'lattice parameter': 4.781,
    'pbe-lp': 4.832,
    'cohesive energy': 5.57,
    'debye temperature': 648.,
    'bulk modulus': 115.,
    'symbols': 'CaO',
    'latex name': "CaO",
    'structure':'rocksalt',
    'magmom': None},
'TiC-b1': {
    'name': 'TiC-b1',
    'lattice parameter': 4.318,
    'pbe-lp': 4.332,
    'cohesive energy': 7.16,
    'debye temperature': 940.,
    'bulk modulus': 233.,
    'symbols': 'TiC',
    'latex name': "TiC",
    'structure':'rocksalt',
    'magmom': None},
'TiN-b1': {
    'name': 'TiN-b1',
    'lattice parameter': 4.226,
    'pbe-lp': 4.247,
    'cohesive energy': 6.69,
    'debye temperature': 757.,
    'bulk modulus': 277.,
    'symbols': 'TiN',
    'latex name': "TiN",
    'structure':'rocksalt',
    'magmom': None},
'ZrC-b1': {
    'name': 'ZrC-b1',
    'lattice parameter': 4.687,
    'pbe-lp': 4.708,
    'cohesive energy': 7.93,
    'debye temperature': 700.,
    'bulk modulus': 207.,
    'symbols': 'ZrC',
    'latex name': "ZrC",
    'structure':'rocksalt',
    'magmom': None},
'ZrN-b1': {
    'name': 'ZrN-b1',
    'lattice parameter': 4.576,
    'pbe-lp': 4.594,
    'cohesive energy': 7.52,
    'debye temperature': 700.,
    'bulk modulus': 240.,
    'symbols': 'ZrN',
    'latex name': "ZrN",
    'structure':'rocksalt',
    'magmom': None},
'VC-b1': {
    'name': 'VC-b1',
    'lattice parameter': 4.148,
    'pbe-lp': 4.154,
    'cohesive energy': 6.94,
    'debye temperature': 971.,
    'bulk modulus': 303.,
    'symbols': 'VC',
    'latex name': "VC",
    'structure':'rocksalt',
    'magmom': None,
    'gpaw-conv': [1.e-6, 1.e-3, 1.e-6]},
'VN-b1': {
    'name': 'VN-b1',
    'lattice parameter': 4.130,
    'pbe-lp': 4.116,
    'cohesive energy': 6.25,
    'debye temperature': 755.,
    'bulk modulus': 268.,
    'symbols': 'VN',
    'latex name': "VN",
    'structure':'rocksalt',
    'magmom': None},
'NbC-b1': {
    'name': 'NbC-b1',
    'lattice parameter': 4.461,
    'pbe-lp': 4.484,
    'cohesive energy': 8.26,
    'debye temperature': 761.,
    'bulk modulus': 300.,
    'symbols': 'NbC',
    'latex name': "NbC",
    'structure':'rocksalt',
    'magmom': None},
'NbN-b1': {
    'name': 'NbN-b1',
    'lattice parameter': 4.383,
    'pbe-lp': 4.422,
    'cohesive energy': 7.50,
    'debye temperature': 730.,
    'bulk modulus': 287.,
    'symbols': 'NbN',
    'latex name': "NbN",
    'structure':'rocksalt',
    'magmom': None},
'FeAl-b2': {
    'name': 'FeAl-b2',
    'lattice parameter': 2.881,
    'pbe-lp': 2.868,
    'cohesive energy': 4.23,
    'debye temperature': 514.,
    'bulk modulus': 136.1,
    'symbols': 'FeAl',
    'latex name': "FeAl",
    'structure':'cesiumchloride',
    'magmom': 0.35},
'CoAl-b2': {
    'name': 'CoAl-b2',
    'lattice parameter': 2.854,
    'pbe-lp': 2.851,
    'cohesive energy': 4.35,
    'debye temperature': 500.,
    'bulk modulus': 162.,
    'symbols': 'CoAl',
    'latex name': "CoAl",
    'structure':'cesiumchloride',
    'magmom': None},
'NiAl-b2': {
    'name': 'NiAl-b2',
    'lattice parameter': 2.881,
    'pbe-lp': 2.892,
    'cohesive energy': 4.46,
    'debye temperature': 402.,
    'bulk modulus': 166.,
    'symbols': 'NiAl',
    'latex name': "NiAl",
    'structure':'cesiumchloride',
    'magmom': None},
'BN-b3': {
    'name': 'BN-b3',
    'lattice parameter': 3.594,
    'pbe-lp': 3.624,
    'cohesive energy': 6.76,
    'debye temperature': 1700.,
    'bulk modulus': 372.3,
    'symbols': 'BN',
    'latex name': "BN",
    'structure':'zincblende',
    'magmom': None},
'BP-b3': {
    'name': 'BP-b3',
    'lattice parameter': 4.527,
    'pbe-lp': 4.548,
    'cohesive energy': 5.14,
    'debye temperature': 985.,
    'bulk modulus': 152.,
    'symbols': 'BP',
    'latex name': "BP",
    'structure':'zincblende',
    'magmom': None},
'BAs-b3': {
    'name': 'BAs-b3',
    'lattice parameter': 4.764,
    'pbe-lp': 4.809,
    'cohesive energy': None,
    'debye temperature': 800.,
    'bulk modulus': 138.,
    'symbols': 'BAs',
    'latex name': "BAs",
    'structure':'zincblende',
    'magmom': None},
'AlN-b3': {
    'name': 'AlN-b3',
    'lattice parameter': 4.368,
    'pbe-lp': 4.402,
    'cohesive energy': 5.85,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'AlN',
    'latex name': "AlN",
    'structure':'zincblende',
    'magmom': None},
'AlP-b3': {
    'name': 'AlP-b3',
    'lattice parameter': 5.450,
    'pbe-lp': 5.504,
    'cohesive energy': 4.32,
    'debye temperature': 588.,
    'bulk modulus': 74.4,
    'symbols': 'AlP',
    'latex name': "AlP",
    'structure':'zincblende',
    'magmom': None},
'AlAs-b3': {
    'name': 'AlAs-b3',
    'lattice parameter': 5.649,
    'pbe-lp': 5.728,
    'cohesive energy': 3.82,
    'debye temperature': 292.,
    'bulk modulus': 74.,
    'symbols': 'AlAs',
    'latex name': "AlAs",
    'structure':'zincblende',
    'magmom': None},
'GaN-b3': {
    'name': 'GaN-b3',
    'lattice parameter': 4.523,
    'pbe-lp': 4.549,
    'cohesive energy': 4.55,
    'debye temperature': 600.,
    'bulk modulus': 195.,
    'symbols': 'GaN',
    'latex name': "GaN",
    'structure':'zincblende',
    'magmom': None},
'GaP-b3': {
    'name': 'GaP-b3',
    'lattice parameter': 5.441,
    'pbe-lp': 5.506,
    'cohesive energy': 3.61,
    'debye temperature': 445.,
    'bulk modulus': 87.4,
    'symbols': 'GaP',
    'latex name': "GaP",
    'structure':'zincblende',
    'magmom': None},
'GaAs-b3': {
    'name': 'GaAs-b3',
    'lattice parameter': 5.641,
    'pbe-lp': 5.751,
    'cohesive energy': 3.34,
    'debye temperature': 344.,
    'bulk modulus': 75.6,
    'symbols': 'GaAs',
    'latex name': "GaAs",
    'structure':'zincblende',
    'magmom': None},
'InP-b3': {
    'name': 'InP-b3',
    'lattice parameter': 5.858,
    'pbe-lp': 5.963,
    'cohesive energy': 3.47,
    'debye temperature': 321.,
    'bulk modulus': 71.,
    'symbols': 'InP',
    'latex name': "InP",
    'structure':'zincblende',
    'magmom': None},
'InAs-b3': {
    'name': 'InAs-b3',
    'lattice parameter': 6.048,
    'pbe-lp': 6.188,
    'cohesive energy': 3.08,
    'debye temperature': 247.,
    'bulk modulus': 58.,
    'symbols': 'InAs',
    'latex name': "InAs",
    'structure':'zincblende',
    'magmom': None},
'SiC-b3': {
    'name': 'SiC-b3',
    'lattice parameter': 4.348,
    'pbe-lp': 4.378,
    'cohesive energy': 6.48,
    'debye temperature': 1232.,
    'bulk modulus': 225.,
    'symbols': 'SiC',
    'latex name': "SiC",
    'structure':'zincblende',
    'magmom': None},
# Pearson
'KBr-b1': {
    'name': 'KBr-b1',
    'lattice parameter': 6.60,
    'revtpss-lp': 6.75894864163,
    'cohesive energy': 3.08,
    'symbols': 'KBr',
    'latex name': "KBr",
    'structure':'rocksalt',
    'magmom': None},
'CaSe-b1': {
    'name': 'CaSe-b1',
    'lattice parameter': 5.80,
    'revtpss-lp': 5.98316194871,
    'cohesive energy': 4.01,
    'symbols': 'CaSe',
    'latex name': "CaSe",
    'structure':'rocksalt',
    'magmom': None},
'SeAs-b1': {
    'name': 'SeAs-b1',
    'lattice parameter': 5.48,
    'revtpss-lp': 5.55199610297,
    'cohesive energy': 2.46, #4.92,
    'symbols': 'SeAs',
    'latex name': "SeAs",
    'structure':'rocksalt',
    'magmom': None},
'RbI-b1': {
    'name': 'RbI-b1',
    'lattice parameter': 7.00,
    'revtpss-lp': None,
    'cohesive energy': 2.71,
    'symbols': 'RbI',
    'latex name': "RbI",
    'structure':'rocksalt',
    'magmom': None},
'LiI-b1': {
    'name': 'LiI-b1',
    'lattice parameter': 6.00,
    'revtpss-lp': 5.98205708599,
    'cohesive energy': 2.78,
    'symbols': 'LiI',
    'latex name': "LiI",
    'structure':'rocksalt',
    'magmom': None},
'CsF-b1': {
    'name': 'CsF-b1',
    'lattice parameter': 6.02,
    'revtpss-lp': 6.21937874326,
    'cohesive energy': 3.66,
    'symbols': 'CsF',
    'latex name': "CsF",
    'structure':'rocksalt',
    'magmom': None},
'CsI-b2': {
    'name': 'CsI-b2',
    'lattice parameter': 4.42,
    'revtpss-lp': 4.6376073499,
    'cohesive energy': 2.75,
    'symbols': 'CsI',
    'latex name': "CsI",
    'structure':'cesiumchloride',
    'magmom': None},
'AgF-b1': {
    'name': 'AgF-b1',
    'lattice parameter': 4.92,
    'revtpss-lp': 4.95954689488,
    'cohesive energy': 2.95,
    'symbols': 'AgF',
    'latex name': "AgF",
    'structure':'rocksalt',
    'magmom': None},
'AgCl-b1': {
    'name': 'AgCl-b1',
    'lattice parameter': 5.56,
    'revtpss-lp': 5.55845581698,
    'cohesive energy': 2.75,
    'symbols': 'AgCl',
    'latex name': "AgCl",
    'structure':'rocksalt',
    'magmom': None},
'AgBr-b1': {
    'name': 'AgBr-b1',
    'lattice parameter': 5.78,
    'revtpss-lp': 5.7885713515,
    'cohesive energy': 2.58,
    'symbols': 'AgBr',
    'latex name': "AgBr",
    'structure':'rocksalt',
    'magmom': None},
'CaS-b1': {
    'name': 'CaS-b1',
    'lattice parameter': 5.70,
    'revtpss-lp': 5.7588414509,
    'cohesive energy': 4.81,
    'symbols': 'CaS',
    'latex name': "CaS",
    'structure':'rocksalt',
    'magmom': None},
'BaO-b1': {
    'name': 'BaO-b1',
    'lattice parameter': 5.52,
    'revtpss-lp': 5.60522427797,
    'cohesive energy': 5.10,
    'symbols': 'BaO',
    'latex name': "BaO",
    'structure':'rocksalt',
    'magmom': None},
'BaSe-b1': {
    'name': 'BaSe-b1',
    'lattice parameter': 6.60,
    'revtpss-lp': 6.67681841117,
    'cohesive energy': 4.03,
    'symbols': 'BaSe',
    'latex name': "BaSe",
    'structure':'rocksalt',
    'magmom': None},
'CdO-b1': {
    'name': 'CdO-b1',
    'lattice parameter': 4.70,
    'revtpss-lp': 4.75039242719,
    'cohesive energy': 3.21,
    'symbols': 'CdO',
    'latex name': "CdO",
    'structure':'rocksalt',
    'magmom': None},
'MnO-b1': {
    'name': 'MnO-b1',
    'lattice parameter': 4.44,
    'revtpss-lp': 4.4520326675,
    'cohesive energy': 4.75,
    'symbols': 'MnO',
    'latex name': "MnO",
    'structure':'rocksalt',
    'magmom': 2.4},
'MnS-b1': {
    'name': 'MnS-b1',
    'lattice parameter': 5.22,
    'revtpss-lp': 5.13039706051,
    'cohesive energy': 4.01,
    'symbols': 'MnS',
    'latex name': "MnS",
    'structure':'rocksalt',
    'magmom': 2.4},
# Haglund
'ScC-b1': {
    'name': 'ScC-b1',
    'lattice parameter': 4.72,
    'revtpss-lp': 4.69537210811,
    'cohesive energy': 6.37,
    'symbols': 'ScC',
    'latex name': "ScC",
    'structure':'rocksalt',
    'magmom': None},
'CrC-b1': { # slow gpaw convergence
    'name': 'CrC-b1',
    'lattice parameter': 4.12,
    'revtpss-lp': 4.12680477339,
    'cohesive energy': 5.80,
    'symbols': 'CrC',
    'latex name': "CrC",
    'structure':'rocksalt',
    'magmom': 0.6},
    #'gpaw-conv': [1.e-5, 1.e-7, 1.e-4]},
'MnC-b1': {
    'name': 'MnC-b1',
    'lattice parameter': 4.12,
    'revtpss-lp': 4.11651335966,
    'cohesive energy': 5.14,
    'symbols': 'MnC',
    'latex name': "MnC",
    'structure':'rocksalt',
    'magmom': 1.2},
    #'gpaw-conv': [1.e-5, 1.e-6, 1.e-4]},
'FeC-b1': {
    'name': 'FeC-b1',
    'lattice parameter': 4.09,
    'revtpss-lp': 3.98592429783,
    'cohesive energy': 5.67,
    'symbols': 'FeC',
    'latex name': "FeC",
    'structure':'rocksalt',
    'magmom': None},
'CoC-b1': {
    'name': 'CoC-b1',
    'lattice parameter': 4.05,
    'revtpss-lp': 3.98558621985,
    'cohesive energy': 5.69,
    'symbols': 'CoC',
    'latex name': "CoC",
    'structure':'rocksalt',
    'magmom': None},
'NiC-b1': {
    'name': 'NiC-b1',
    'lattice parameter': 3.99,
    'revtpss-lp': 4.03667215219,
    'cohesive energy': 5.65,
    'symbols': 'NiC',
    'latex name': "NiC",
    'structure':'rocksalt',
    'magmom': None},
'ScN-b1': {
    'name': 'ScN-b1',
    'lattice parameter': 4.51,
    'revtpss-lp': 4.52753210905,
    'cohesive energy': 6.72,
    'symbols': 'ScN',
    'latex name': "ScN",
    'structure':'rocksalt',
    'magmom': None},
'CrN-b1': {
    'name': 'CrN-b1',
    'lattice parameter': 4.15,
    'revtpss-lp': 4.18609030986,
    'cohesive energy': 5.14,
    'symbols': 'CrN',
    'latex name': "CrN",
    'structure':'rocksalt',
    'magmom': 1.3},
'MnN-b1': {
    'name': 'MnN-b1',
    'lattice parameter': 4.20,
    'revtpss-lp': 4.15891205836,
    'cohesive energy': 4.08,
    'symbols': 'MnN',
    'latex name': "MnN",
    'structure':'rocksalt',
    'magmom': 1.6},
    #'gpaw-conv': [1.e-5, 1.e-7, 1.e-4]},
'FeN-b1': {
    'name': 'FeN-b1',
    'lattice parameter': 4.13,
    'revtpss-lp': 4.12,
    'cohesive energy': 4.59,
    'symbols': 'FeN',
    'latex name': "FeN",
    'structure':'rocksalt',
    'magmom': 1.3},
    #'gpaw-conv': [1.e-5, 1.e-4, 1.e-7]},
'CoN-b1': {
    'name': 'CoN-b1',
    'lattice parameter': 4.10,
    'revtpss-lp': 3.99392851055,
    'cohesive energy': 4.53,
    'symbols': 'CoN',
    'latex name': "CoN",
    'structure':'rocksalt',
    'magmom': None},
'NiN-b1': {
    'name': 'NiN-b1',
    'lattice parameter': 4.10,
    'revtpss-lp': 4.03914234139,
    'cohesive energy': 4.48,
    'symbols': 'NiN',
    'latex name': "NiN",
    'structure':'rocksalt',
    'magmom': None},
'MoC-b1': {
    'name': 'MoC-b1',
    'lattice parameter': 4.278,
    'revtpss-lp': 4.39395801553,
    'cohesive energy': 7.22,
    'symbols': 'MoC',
    'latex name': "MoC",
    'structure':'rocksalt',
    'magmom': None},
'RuC-b1': {
    'name': 'RuC-b1',
    'lattice parameter': 4.129,
    'revtpss-lp': 4.33092551761,
    'cohesive energy': 6.73,
    'symbols': 'RuC',
    'latex name': "RuC",
    'structure':'rocksalt',
    'magmom': None},
'RhC-b1': {
    'name': 'RhC-b1',
    'lattice parameter': 4.145,
    'revtpss-lp': 4.35180356284,
    'cohesive energy': 6.23,
    'symbols': 'RhC',
    'latex name': "RhC",
    'structure':'rocksalt',
    'magmom': None},
'PdC-b1': {
    'name': 'PdC-b1',
    'lattice parameter': 4.221,
    'revtpss-lp': 4.4101528149,
    'cohesive energy': 5.36,
    'symbols': 'PdC',
    'latex name': "PdC",
    'structure':'rocksalt',
    'magmom': None},
'MoN-b1': {
    'name': 'MoN-b1',
    'lattice parameter': 4.214,
    'revtpss-lp': 4.37920656361,
    'cohesive energy': 6.20,
    'symbols': 'MoN',
    'latex name': "MoN",
    'structure':'rocksalt',
    'magmom': None},
'RuN-b1': {
    'name': 'RuN-b1',
    'lattice parameter': 4.058,
    'revtpss-lp': 4.34894810722,
    'cohesive energy': 5.02,
    'symbols': 'RuN',
    'latex name': "RuN",
    'structure':'rocksalt',
    'magmom': None},
'RhN-b1': {
    'name': 'RhN-b1',
    'lattice parameter': 4.082,
    'revtpss-lp': 4.35633367368,
    'cohesive energy': 4.78,
    'symbols': 'RhN',
    'latex name': "RhN",
    'structure':'rocksalt',
    'magmom': None},
'PdN-b1': {
    'name': 'PdN-b1',
    'lattice parameter': 4.145,
    'revtpss-lp': 4.41700841872,
    'cohesive energy': 4.03,
    'symbols': 'PdN',
    'latex name': "PdN",
    'structure':'rocksalt',
    'magmom': None},
'LaC-b1': {
    'name': 'LaC-b1',
    'lattice parameter': 5.429,
    'revtpss-lp': 5.50510598304,
    'cohesive energy': 5.74,
    'symbols': 'LaC',
    'latex name': "LaC",
    'structure':'rocksalt',
    'magmom': None},
'TaC-b1': {
    'name': 'TaC-b1',
    'lattice parameter': 4.457,
    'revtpss-lp': 4.48669679096,
    'cohesive energy': 8.56,
    'symbols': 'TaC',
    'latex name': "TaC",
    'structure':'rocksalt',
    'magmom': None},
'WC-b1': {
    'name': 'WC-b1',
    'lattice parameter': 4.266,
    'revtpss-lp': 4.40260619124,
    'cohesive energy': 8.25,
    'symbols': 'WC',
    'latex name': "WC",
    'structure':'rocksalt',
    'magmom': None},
'OsC-b1': {
    'name': 'OsC-b1',
    'lattice parameter': 4.176,
    'revtpss-lp': 4.35787411212,
    'cohesive energy': 7.36,
    'symbols': 'OsC',
    'latex name': "OsC",
    'structure':'rocksalt',
    'magmom': None},
'IrC-b1': {
    'name': 'IrC-b1',
    'lattice parameter': 4.129,
    'revtpss-lp': 4.39419683105,
    'cohesive energy': 6.84,
    'symbols': 'IrC',
    'latex name': "IrC",
    'structure':'rocksalt',
    'magmom': None},
'PtC-b1': {
    'name': 'PtC-b1',
    'lattice parameter': 4.206,
    'revtpss-lp': 4.45741920763,
    'cohesive energy': 6.34,
    'symbols': 'PtC',
    'latex name': "PtC",
    'structure':'rocksalt',
    'magmom': None},
'LaN-b1': {
    'name': 'LaN-b1',
    'lattice parameter': 5.305,
    'revtpss-lp': 5.35277831785,
    'cohesive energy': 6.27,
    'symbols': 'LaN',
    'latex name': "LaN",
    'structure':'rocksalt',
    'magmom': None},
'TaN-b1': {
    'name': 'TaN-b1',
    'lattice parameter': 4.340,
    'revtpss-lp': 4.44437817766,
    'cohesive energy': 7.63,
    'symbols': 'TaN',
    'latex name': "TaN",
    'structure':'rocksalt',
    'magmom': None},
'WN-b1': {
    'name': 'WN-b1',
    'lattice parameter': 4.202,
    'revtpss-lp': 4.39478109563,
    'cohesive energy': 7.01,
    'symbols': 'WN',
    'latex name': "WN",
    'structure':'rocksalt',
    'magmom': None},
'OsN-b1': {
    'name': 'OsN-b1',
    'lattice parameter': 4.058,
    'revtpss-lp': 4.37261212422,
    'cohesive energy': 5.61,
    'symbols': 'OsN',
    'latex name': "OsN",
    'structure':'rocksalt',
    'magmom': None},
'IrN-b1': {
    'name': 'IrN-b1',
    'lattice parameter': 4.074,
    'revtpss-lp': 4.40354481322,
    'cohesive energy': 5.13,
    'symbols': 'IrN',
    'latex name': "IrN",
    'structure':'rocksalt',
    'magmom': None},
'PtN-b1': {
    'name': 'PtN-b1',
    'lattice parameter': 4.137,
    'revtpss-lp': 4.48127067537,
    'cohesive energy': 4.63,
    'symbols': 'PtN',
    'latex name': "PtN",
    'structure':'rocksalt',
    'magmom': None},
}



from dbgen.support.datatypes.table      import Table,Col,FK
from dbgen.support.datatypes.rule       import Rule
from dbgen.support.datatypes.sqltypes   import Varchar,Decimal

##########################################################################################
##########################################################################################
##########################################################################################

model = Table('model'
    ,desc = 'linear regression models'
    ,cols = [Col('id',            pk = True,auto = True)
            ,Col('name',Varchar(),nn = True)])

hyperparameter = Table('hyperparameter'
    ,desc='List of triplestore values (FK''d from hyperparameter_value)'
    ,cols = [Col('hyperparameter_id',pk = True,auto = True)
            ,Col('name',Varchar())])

hyperparameter_value = Table('hyperparameter_value'
    ,desc='Triplestore table describing hyperparameters for a machine learning application'
    ,cols = [Col('hyperparameter_id',pk = True,auto = True)
            ,Col('name', Varchar(),  nn = True,uniq = True)
            ,Col('value',Varchar())]
    ,fks = [FK('hyperparameter_id','hyperparameter')
            ,FK('name','hyperparameter_type')])

hyperparameter_type = Table('hyperparameter_type'
    ,desc="We'll be dogmatic here and have a universal mapping between hyperparameter names and datatypes"
    ,cols = [Col('name',    Varchar(), pk = True)
            ,Col('datatype',Varchar(), nn = True)])

struct_dataset = Table('struct_dataset'
    ,desc = 'Data to be used for machine learning that is associated with a chemical species'
    ,cols = [Col('struct_dataset_id',pk = True, auto = True)
            ,Col('name',Varchar(),   nn = True, uniq = True)])

struct_dataset_element = Table('struct_dataset_element'
    ,desc ='A fact about some chemical species from a dataset'
    ,cols = [Col('struct_dataset_id',           pk = True)
            ,Col('species_id',                  pk = True)
            ,Col('property',        Varchar(),  pk = True)
            ,Col('value',           Decimal(),  nn = True)]
    ,fks  = [FK('struct_dataset_id','struct_dataset')
            ,FK('species_id',       'species')])

rxn_dataset  = Table('rxn_dataset'
    ,desc = 'Data to be used for machine learning that is associated with a chemical species'
    ,cols = [Col('rxn_dataset_id',  pk = True,auto = True)
            ,Col('name',Varchar(),  nn = True,uniq = True)])

rxn_dataset_element = Table('rxn_dataset_element'
    ,desc = 'A fact about some chemical species from a dataset'
    ,cols = [Col('rxn_dataset_id',          pk = True)
            ,Col('rxn_id',                  pk = True)
            ,Col('property',    Varchar(),  nn = True)
            ,Col('value',       Decimal(),  nn = True)]
    ,fks  = [FK('rxn_dataset_id','rxn_dataset')
            ,FK('rxn_id','rxn')])

tables = [hyperparameter
         ,hyperparameter_type
         ,hyperparameter_value
         ,struct_dataset
         ,struct_dataset_element
         ,rxn_dataset
         ,rxn_dataset_element]
##########################################################################################
##########################################################################################
##########################################################################################

rules = [] # type: list
